﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Controller : MonoBehaviour
{

    /*
     
                                          MANUAL OF THE KINEMATIC, PID CONTROLLER AND TRAJECTORY UPLOAD FUNCTIONS
    
    ------------------------------- KINEMATICS -------------------------------
    This script allows to simulate the dynamic of a four wheeled car.
    The kinematic of the vehicle is completely general:
    each wheel can be both motor and steering
    to set such behaviour it's necessary to tick the box from
    Unity in the tab of the script attached to the car object.
    Good results with this scripts were obtained in test for the car-like front steering
    rear traction model and for the differential tricycle-like kinematic model.

    Each wheel collider is identified by in Unity a unique flag (look for Tag in the inspector of each wheel ), namely:
    Front_right
    Front_left
    Rear_right
    Rear_left

    In order to access them in this script such flags MUST be attached.
    To simulate the dynamic of the wheels it's enough to modify the values of the applied torque
    an the steering angle in this Controller script.
    The Torque is the value [Nm] applied to the wheel with respect to the rotary axis.
    The Steering angle is the angle [°] that the wheel has with respect to the local reference
    frame attached to the chassis. A positive angle applied to the front wheels will make the car
    turn right and a negative value to turn to the left.

    How to modify the value of steering angle and torque: EXAMPLE OF FRONT LEFT WHEEL
    steeringAngle   ---->    Kinematic.SteeringFrontLeft = float Value;
    MotorTorque     ---->    Kinematic.TorqueFrontLeft = float Value;
    
    To specify the wheel on which change the value.
    
    Please notice that not every possible assigned values set is physically sound.
    Be aware that in some cases kinematic singularities are possible, and this may 
    result in undesirable results in the simulation.
    That may happen for instance if the wheels point towards different directions, an thus the vehicle
    becomes disarticulated. Or else, if the angle between the front and the rear wheels is close to the
    value of 90°, in a rear traction vehicle, then the front wheels drag force will
    be very high: depending on the torque provided to the vehicle either the motion will be impossible or
    else the vehicle will move with great vibrations and unstable motion.
    To avoid this it's sufficient to properly set the constraints between the wheels, for example
    impose the same value to the front steering angle in a front steering car-like or use a more
    sophisticated Ackermann steering algorithm, after calculating the istantaneous center of rotation.
    Moreover it's advisable to set a maximum steering angle and torque value.

    If for some particular reason another wheel was to be added:
        
    If it's idle wheel with fixed axis or freely rotating one can attach a wheel collider to the main body
    through a proper unity constraint. For example a caster wheel can be obtained with
    revolute joint and a wheel collider attached to the car body (and not to the visual body).

    If it's a controlled steering or motor wheel instead, please follow these steps:
    tag it with a proper non redundant name in Unity, for example "New_wheel". Then in the Kinematic script add two new
    variables of type float for the desired torque and steering angle.

    public static float newTorque;
    public static float newSteering;
        
    Remember to change the values in the axle object, the tagged object will be selected 
    and their steering angle changed in accordance to the value of newSteering and newTorque

    foreach (AxleInfo axleInfo in axleInfos)
    {
        if (axleInfo.steering)
        {
            switch (axleInfo.Wheel.tag)
            {
                ...

                case "New_wheel":
                    axleInfo.Wheel.steerAngle = newSteering;
                    break;
            }
                       
        }

        if (axleInfo.motor)
        {
            switch (axleInfo.Wheel.tag)
            {
                ...

                case "New_wheel":
                    axleInfo.Wheel.motorTorque = newTorque;
                    break;
            }

        }

    Then it will be sufficient to change the value of Kinematics.newSteering and Kinematics.newTorque
    from the Controller script along the values of the other wheels in the Control script.


    ------------------------------- PID CONTROLLER  -------------------------------

    If needed, a PID function is provided:
    such function, given a target, for instance the desired forward velocity
    of the vehicle, provides the Torque to the wheel in order to achieve it.
    In order to the algorithm to work, the actual velocity of the vehicle, must be
    passed to the function. From that, by difference, the error from
    the target is calculated.
    The output torque is a float value obtained with the sum of three components:
    - proportional to the error, with constant Kp
    - proportional to the derivative of the error, with constant Kd
    - proportional to the integral of the error, with constant Ki

    In the example below it is possible to see how it works.

    
    ------------------------------- TRAJECTORY UPLOAD -------------------------------

    You will use the Init() function like in the following:

    Vector3[] NameTrajectory = ReadTrajectory.Init(filename);
        
    The filename can be changed in the Unity console directly in the inspector. 
    If more than one file needs to be imported, one can simply add more filenames in the initialization.
    This filepath must coincide with the root path of the project, for example
    if you saved your project as "MyProject", then the file to be read must be in
    "./MyProject/" not in the Script subfolder nor in any other subfolder.
    The format of the vector must divided by commas, while decimal separators are points.
    (xValue, yValue, zValue)
    The format of each line can either be in parentheses or not.
        
    */

    // ---------------------------------------------------- Example of a Car-Like controlled in velocity ------------------------------------------


    // Define the parameters needed in the script.

    //Definition of a rigidbody. It is needed since we need to access the velocity of such object.
    //In our case attach the car_root to that rigidbody in the unity inspector.
    public Rigidbody movingBody;
    //Vector needed to store the car velocity.
    Vector3 ActualVelocity;
    //PID parameters
    public float Kp = 100, Ki = 0, Kd = 0, PastVelocity= 0;
    float referenceValue;
    float intError = 0;
    //Definition of the file where the trajectory is saved in.
     public string filename = "VectorExample.txt";


    //Definition of the PID controller function
    float PIDControl(float actualValue, float referenceValue, float pastValue = 0)
    {
        float output;
        intError += (referenceValue - actualValue) * Time.fixedDeltaTime;
        output = Kp * (referenceValue - actualValue) + Kd * (actualValue - pastValue) / Time.fixedDeltaTime + Ki * intError;
        return output;
    }

    // Call the start function to initialize 
    void Start()
    {
        // EXAMPLE OF TEXT READ AND DEBUG ON CONSOLE
            Vector3[] example = ReadTrajectory.Init(filename);
            for (int i = 0; i < example.Length; i ++)
            {
                 Debug.Log(example[i]);
            }


        // INITIALIZE THE REFERENCE VELOCITY FOR THE PID CONTROLLER
            referenceValue = 3;
    }

    
    // Update is called once per frame
    void FixedUpdate () {


        // Use the method Rigidbody.velocity to get the actual velocity of the car
        movingBody = GetComponent<Rigidbody>();
        ActualVelocity = transform.InverseTransformDirection(movingBody.velocity);

        //Modify the Torque on the rear wheels in order to have actualVelocity equal to referenceValue
        Kinematic.TorqueRearLeft = PIDControl(ActualVelocity.z, referenceValue, PastVelocity);
        Kinematic.TorqueRearRight = Kinematic.TorqueRearLeft;
        Kinematic.SteeringFrontLeft = 10;
        Kinematic.SteeringFrontRight = 10;
        PastVelocity = ActualVelocity.z;
        Debug.Log(ActualVelocity.z);
        
    }
}
